// BMS16v3: 8-16 Cell Integrated Battery Management System
// Open Source version, released under MIT License (see Readme file)
// Last modified by Ian Hooper (ZEVA), August 2021

// Code for ATmega16M1 (also suitable for ATmega32M1, ATmega64m1)
// Fuses: 8Mhz+ external crystal with 0ms startup time, CKDIV8 off, brownout 4.2V

// Note: A few #defines for build options found in Common.h file too

#define BMS_VERSION 3	// (v)2 or (v)3 only, to accommodate a few minor changes between these versions
#define IS_BMS12I	0	// Same code is used for BMS12i as well with minor changes, set to 0 for BMS16, 1 for BMS12i

#define LOWER_LTC_CORRECTION	6 // Millivolts added to compensate for voltage loss in buffer resistors and LTC6802 variation
#define UPPER_LTC_CORRECTION	6 // This second one applies to BMS16 only (top 8 cells). Both are 6 average but may vary +/-3

#define POWER_OFF_MILLIVOLTS 2000 // Normally 2000
#define SHUNT_CALIBRATION	103L	// % calibration/compensation, often needs to be 103L for BMS16, 100L for BMS12i
#define ENABLE_MONITOR	1
#define SMA_INTEGRATION	1 // 500kbps 11-bit CAN - also used for Pylontech
#define SOLAX_INTEGRATION 0 // 500kbps 11-bit CAN - NOT TESTED SUCCESSFULLY, needs work
#define INVERT_SSRS		0 // For negative logic, outputs closed for error - warn customers about losing "fail safe"
#define CURRENT_OFFSET 0L // For the rare cases where automatic calibration isn't sufficient
#define USE_600A_SHUNT	0 // Was used briefly when I had some 600A shunts (instead of 500A)
#define LOCK_SETTINGS	0 // Can prevent settings from being changed if desired

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>

#include "Common.h" // Note - this file is shared with several other products, best to use a single copy somewhere

#define true 1
#define false 0

enum { NO_SHUNT, SHUNT_100A, SHUNT_200A, SHUNT_500A };

// SPI interface pins (for comms with LTC6802 voltage samplers)
#define	CSBI		(1<<PB3)
#define CSBI_PORT	PORTB
#define	SDO			(1<<PC6)
#define SDO_PORT	PINC
#define SDI			(1<<PB4)
#define SDI_PORT	PORTB
#define SCKI		(1<<PB5)
#define SCKI_PORT	PORTB

#define CSBI2		(1<<PB2)
#define CSBI2_PORT	PORTB
#define SDO2		(1<<PD7)
#define SDO2_PORT	PIND
#define SDI2		(1<<PD4)
#define SDI2_PORT	PORTD
#define SCKI2		(1<<PB1)
#define SCKI2_PORT	PORTB

// General I/O
#define LV_RELAY	(1<<PD5)
#define HV_RELAY	(1<<PD6)
#define RELAY_PORT	PORTD

#define PWR_HOLD	(1<<PD0)
#define ENABLE12V	(1<<PB6)
#define DISABLE_CAN	(1<<PB0)
#define SLEEP		!(PINC & (1<<PC0)) // Pull down to enable sleep

// Status LED
#define LED_PORT	PORTD
#define GREEN		(1<<PD3)
#define RED			(1<<PD2)

// ADCs
#define TEMP		4
#define	POS			9
#define NEG			8

// LTC6802 Command codes
#define WRCFG	0x01
#define RDCFG	0x02
#define	RDCV	0x04
#define RDTMP	0x08
#define STCVAD	0x10
#define STTMPAD	0x30

// Function declarations
void SetupPorts();
void CanTX(long packetID, char bytes);
void WriteSPIByte(unsigned char byte);
unsigned char ReadSPIByte();
void WriteSPIByte2(unsigned char byte);
unsigned char ReadSPIByte2();
char LoadSettingsFromEEPROM();
void SaveSettingsToEEPROM();
long GetCurrent();
int GetTemperature();

// Global variables
uint8_t txData[8];
volatile int voltage[16]; // In millivolts
char error;
unsigned long errorAckFlags = 0;
char soc;
long charge, maxCharge;
long current;
int canCurrentSensorTimeout = 0;
unsigned char lvRelayEnabled = false, hvRelayEnabled = false;
volatile short ticks = 0; // 128Hz polling
short ticks2 = 0; // 16Hz polling
char sleeping = false;
char justWokeUp = true;

// ADC vals for temps:  0   20   40   60   80 100 120 140 160 dec C
short tempData[9] = { 783, 568, 356, 204, 114, 65, 38, 23, 15 };
short tempBuffer;
short temperature;

int shuntCalibrationFine = 0;
int shuntCalibrationCoarse = 0;
char calibrateCurrentRequest = false;
char saveToEepromFlag = false;

SIGNAL(CAN_INT_vect) // Interrupt function called when a CAN message is received
{
	int8_t length, savedCANPage;

	unsigned char data[8];

	savedCANPage = CANPAGE; // Saves current MOB
	CANPAGE = CANHPMOB & 0xF0; // Selects MOB with highest priority interrupt
	if (CANSTMOB & (1<<RXOK))
	{
		length = CANCDMOB & 0x0F; // Number of bytes to receive is bottom four bits of this reg
		for (int8_t i=0; i<length; i++)
			data[i] = CANMSG; // This autoincrements when read
		
		// ID has top 8 bits in IDT1 and bottom 3 bits at TOP of IDT2
		long packetID;
		if (USE_29BIT_IDS)
			packetID = (((long)CANIDT1)<<21) + (((long)CANIDT2)<<13) + (CANIDT3<<5) + (CANIDT4>>3);
		else
			packetID = (CANIDT1<<3) + (CANIDT2>>5);

		if (packetID == CORE_ACKNOWLEDGE_ERROR)
		{
			if (data[0] < NUM_ERRORS) errorAckFlags |= (1L<<data[0]);
			error = NO_ERROR;
		}
		else if (packetID == CORE_RECEIVE_CONFIG1)
		{
			if (!LOCK_SETTINGS)
			{
				for (int n=0; n<8; n++) settings[n] = Cap(data[n], minimums[n], bms16maximums[n]);
				if (IS_BMS12I) settings[NUM_CELLS] = Cap(data[NUM_CELLS], 4, 12); // Different min/max cells for BMS12i
			}
		}
		else if (packetID == CORE_RECEIVE_CONFIG2)
		{
			if (!LOCK_SETTINGS)
				for (int n=0; n<8; n++) settings[n+8] = Cap(data[n], minimums[n+8], bms16maximums[n+8]);
		}
		else if (packetID == CORE_RECEIVE_CONFIG3)
		{
			if (!LOCK_SETTINGS)
				for (int n=0; n<8; n++) settings[n+16] = Cap(data[n], minimums[n+16], bms16maximums[n+16]);
		}
		else if (packetID == CORE_RECEIVE_CONFIG4)
		{
			if (!LOCK_SETTINGS)
			{
				for (int n=0; n<8; n++) settings[n+24] = Cap(data[n], minimums[n+24], bms16maximums[n+24]);
				saveToEepromFlag = true; // Only do this after 4th packet is received
			}
		}
		else if (packetID == CORE_RESET_SOC)
		{
			charge = maxCharge;
		}
		else if (packetID == CAN_ZERO_CURRENT)
		{
			calibrateCurrentRequest = true;
		}
		else if (packetID == CAN_CURRENT_SENSOR_ID && settings[SHUNT_SIZE] == NO_SHUNT)
		{
			current = ((long)data[2]) + ((long)data[1]<<8) + ((long)data[0]<<16) - 8388608L;
			canCurrentSensorTimeout = 4; // 1 second timeout in 4hz loop
		}
		else if (packetID == POWER_OFF)
		{
			PutLTCsToSleep();
			PORTD &= ~PWR_HOLD;
		}

		CANCDMOB = (1<<CONMOB1) | (8<<DLC0) | ((1<<IDE)*USE_29BIT_IDS); // Enable reception, 11-bit IDE, data length 8
		// Note: The DLC field of CANCDMOB register is updated by the received MOB, and if it differs from above, an error is set
	}
	CANSTMOB = 0x00; // Reset interrupt reason on selected channel
	CANPAGE = savedCANPage;
}

SIGNAL(TIMER0_OVF_vect) // Called at 31250Hz Hz
{
	ticks++; // Used for main loop timing
}

SIGNAL(INT3_vect)
{
	// Interrupt function called when waking from sleep, needs to be here but does nothing
}

// Analog to digital function
#define ADC_VREF_TYPE (1<<REFS0) // AVCC ref with cap on AREF pin
int ReadADC(unsigned char channel)
{
	ADMUX = channel | ADC_VREF_TYPE;
	_delay_us(10);				// Delay needed for the stabilization of the ADC input voltage
	ADCSRA |= (1<<ADSC);				// Start the AD conversion
	while ((ADCSRA & 0x10) == 0);	// Wait for the AD conversion to complete
	ADCSRA |= 0x10;
	return ADCW;
}

void ResetMaxCharge()
{
	maxCharge = (long)settings[PACK_CAPACITY]*PACK_CAPACITY_MULTIPLIER*360000L;
}

void CalibrateShunt()
{
	wdt_disable();

	shuntCalibrationCoarse = 0;
	shuntCalibrationFine = 0;

	// Initialise shunt calibration
	for (int n=0; n<20; n++)
	{
		shuntCalibrationCoarse += (ReadADC(POS) - ReadADC(NEG))*10L;
		_delay_ms(10);
	}
	for (int n=0; n<20; n++)
	{
		int fine = ReadADC(15);
		if (fine > 511) fine -= 1024; // Make it signed
		shuntCalibrationFine += fine;
		_delay_ms(10);
	}
	shuntCalibrationFine /= 20;
	shuntCalibrationCoarse /= 20;
}

void PutLTCsToSleep()
{
	CSBI_PORT &= ~CSBI; // Pull down to start command
	WriteSPIByte(WRCFG); // write configuration group
	for (int n=0; n<6; n++) WriteSPIByte(0b00000000); // Zeroes in the configuration register group
	CSBI_PORT |= CSBI; // Pull up to end command

	if (!IS_BMS12I) // BMS16 has two LTC chips
	{
		CSBI2_PORT &= ~CSBI2; // Pull down to start command - second LTC
		WriteSPIByte2(WRCFG); // write configuration group
		for (int n=0; n<6; n++) WriteSPIByte2(0b00000000); // Zeroes in the configuration register group
		CSBI2_PORT |= CSBI2; // Pull up to end command
	}
}

void main()
{
	SetupPorts();
	_delay_ms(200); // Let things settle

	PORTD |= PWR_HOLD;
	PORTB |= ENABLE12V;

	char result = LoadSettingsFromEEPROM();
	if (result == EEPROM_BLANK || result == EEPROM_CORRUPT)
	{
		error = CORRUPT_EEPROM_ERROR;
		CalibrateShunt();
		SaveSettingsToEEPROM();
	}	

	ResetMaxCharge();
	charge = maxCharge/2L; // Startup with 50% SoC - can't be sure upon power up, so assume middle (allows SMAs to still charge new packs)

	// Initialising variables
	for (int n=0; n<8; n++) txData[n] = 0;
	for (int n=0; n<16; n++) voltage[n] = 0;

	sei(); // Enable interrupts
	wdt_enable(WDTO_250MS);

	unsigned char cellBytes[25];
	unsigned char cellBytes2[25];
	short voltages[16][8];

	long currentBuffer[4] = { 0, 0, 0, 0 }; // Init to zero
	char currentCounter = 0;
	short tempBuffer[4] = { 0, 0, 0, 0 };

	short lvRelayTimer = 0;
	short hvRelayTimer = 0;

	short criticalVoltageTimer = 0;

	long packVoltage;
	int midpointCellVoltage = 5000; // Used for balancing so initialise at high millivolts

	short overcurrentTimer = 0;
	char overcurrentTrip = false;
	char thermalTrip = false;
	char tooColdToCharge = false;

	unsigned char counter = 0;
	char slowCounter = 0;
	unsigned char smaPacket = 0;
	unsigned char solaxPacket = 0;
	while (1)
	{
		wdt_reset();

		if (calibrateCurrentRequest)
		{
			CalibrateShunt();
			SaveSettingsToEEPROM(); // Easiest way to save calibration too
			calibrateCurrentRequest = false;
		}

		if (saveToEepromFlag)
		{
			saveToEepromFlag = false;
			SaveSettingsToEEPROM();
			ResetMaxCharge();
		}

		if (ticks > 244) // 128Hz
		{
			ticks -= 244;
			ticks2++;

			currentBuffer[counter] += GetCurrent();

			if (ticks2 == 4) // Start sampling LTCs because we need to give them 20ms to finish
			{
				// Comms with LTC6802...

				int averageCellVoltage = packVoltage / (long)settings[NUM_CELLS];
				int shuntVoltage = 2000+settings[BALANCE_VOLTAGE]*10;
				if (settings[BALANCE_VOLTAGE] == 251) // Dynamic shunt voltage
					shuntVoltage = midpointCellVoltage+BALANCE_TOLERANCE;
					// Oct 2020: Changed averageCellVoltage to midpointCellVoltage for new balancing scheme

				// Calculate shunt bits, LTC #1
				unsigned short shuntBits = 0;
				if (!sleeping)
				{
					int numCellsOnFirstLTC = 8; // 30 May 2019: Couple of new lines to fix bug of shunts applied to wrong cells for < 12 cell packs
					if (IS_BMS12I)
						numCellsOnFirstLTC = settings[NUM_CELLS];
					else if (settings[NUM_CELLS] < 12)
						numCellsOnFirstLTC = settings[NUM_CELLS] - 4;
					for (int n=0; n<numCellsOnFirstLTC; n++)
						if (voltage[n] > shuntVoltage) shuntBits += (1<<n);
				}

				// Write config registers, LTC #1
				CSBI_PORT &= ~CSBI; // Pull down to start command
				WriteSPIByte(WRCFG); // write configuration group
				WriteSPIByte(0b00000001);
				WriteSPIByte(((unsigned short)(shuntBits & 0x00FF))); // Bottom byte of shunt bits
				WriteSPIByte(((unsigned short)(shuntBits>>8))); // Top four bits of shunt bits, right shifted 1 byte
				WriteSPIByte(0b00000000);
				WriteSPIByte(0b00000000);
				WriteSPIByte(0b00000000);	
				CSBI_PORT |= CSBI; // Pull up to end command
				_delay_us(100);

				if (!IS_BMS12I) // BMS16 has second LTC to talk to
				{
					// Write config registers, LTC #2
					shuntBits = 0; // Prepare shunt bits
					if (!sleeping)
					{
						int numCellsOffset = 8; // 30 May 2019: A few new lines to fix bug of shunts applied to wrong cells for < 12 cell packs
						if (settings[NUM_CELLS] < 12) // then we always have 4 cells on this LTC but at varying potential
							numCellsOffset = settings[NUM_CELLS]-4;
						for (int n=0; n<8; n++)
							if (voltage[n+numCellsOffset] > shuntVoltage) shuntBits += (1<<n);
					}
						
					CSBI2_PORT &= ~CSBI2; // Pull down to start command
					WriteSPIByte2(WRCFG); // write configuration group
					WriteSPIByte2(0b00000001);
					WriteSPIByte2(((unsigned short)(shuntBits & 0x00FF))); // Bottom byte of shunt bits
					WriteSPIByte2(((unsigned short)(shuntBits>>8))); // Top four bits of shunt bits, right shifted 1 byte
					WriteSPIByte2(0b00000000);
					WriteSPIByte2(0b00000000);
					WriteSPIByte2(0b00000000);	
					CSBI2_PORT |= CSBI2; // Pull up to end command
					_delay_us(100);
				}

				// Start voltage sampling ,LTC #1
				CSBI_PORT &= ~CSBI;
				WriteSPIByte(STCVAD);
				CSBI_PORT |= CSBI;

				if (!IS_BMS12I)
				{
					// Start voltage sampling, lTC #2
					CSBI2_PORT &= ~CSBI2;
					WriteSPIByte2(STCVAD);
					CSBI2_PORT |= CSBI2;
				}
			}
		}

		if (ticks2 > 8) // 16Hz loop
		{
			ticks2 -= 8;
	
			// Read cell voltage registers, LTC #1
			CSBI_PORT &= ~CSBI;
			WriteSPIByte(RDCV);
			SDI_PORT |= SDI;
			for (int n=0; n<25; n++) cellBytes[n] = ReadSPIByte();
			CSBI_PORT |= CSBI;
			_delay_us(100);

			// Extract voltage data, LTC #1
			int numCells = 8;
			if (IS_BMS12I) numCells = 12;
			for (int n=0; n<numCells; n+=2)
			{
				voltages[n][counter] = (cellBytes[n*3/2] + 256 * (cellBytes[n*3/2+1] & 0x0F))*3/2;
				voltages[n+1][counter] = (((cellBytes[n*3/2+1] & 0xF0)>>4) + cellBytes[n*3/2+2]*16)*3/2;
			}
	
			if (!IS_BMS12I)
			{
				// Read cell voltage registers, LTC #2
				CSBI2_PORT &= ~CSBI2;
				WriteSPIByte2(RDCV);
				SDI2_PORT |= SDI2;
				for (int n=0; n<25; n++) cellBytes2[n] = ReadSPIByte2();
				CSBI2_PORT |= CSBI2;
				_delay_us(100);

				// Extract voltage data, LTC #2
				char offset = 8;
				if (settings[NUM_CELLS] < 12) // Then we have to use fewer of the first LTC's cells, overwrite some..
					offset = 8 + settings[NUM_CELLS] - 12;
				for (int n=0; n<8; n+=2)
				{
					voltages[n+offset][counter] = (cellBytes2[n*3/2] + 256 * (cellBytes2[n*3/2+1] & 0x0F))*3/2;
					voltages[n+offset+1][counter] = (((cellBytes2[n*3/2+1] & 0xF0)>>4) + cellBytes2[n*3/2+2]*16)*3/2;
				}
			}

			tempBuffer[counter] = GetTemperature();
			temperature = 0;
			for (int n=0; n<4; n++) temperature += tempBuffer[n];
			temperature /= 4;

			char wasSleeping = sleeping;
			sleeping = SLEEP;
			if (wasSleeping && !sleeping)
			{
				errorAckFlags = 0; // Reset all ack flags e.g so we get Low Cell warnings again
				overcurrentTrip = false;
				justWokeUp = true;
			}
			
			if (sleeping)
			{
				PORTB &= ~ENABLE12V;
				PORTB |= DISABLE_CAN;
			}
			else 
			{
				PORTB |= ENABLE12V;
				PORTB &= ~DISABLE_CAN;
			
				if (ENABLE_MONITOR)
				{
					if (counter == 0) // Sending data to Monitor, iterate through 4 different packets at 16Hz
					{
						unsigned short ampHours = charge / 36000L; // It's actually in tenths of amp hours
						unsigned short v = packVoltage / 100L; // i.e Tenths of a volt					

						// EVMS3 packet format
						char status = RUNNING;
						if (!lvRelayEnabled || !hvRelayEnabled) status = STOPPED;
						txData[0] = status + (error<<3);	// Status in bottom 3 bits, error top 5
						txData[1] = ampHours>>8;
						txData[2] = ampHours&0xFF;
						txData[3] = v>>8;
						txData[4] = v&0xFF;
						txData[5] = 255; // Sending 255 for auxVoltage will tell the Monitor that we're a BMS16v2
						txData[6] = 0;//isolation + (headlightsOn<<7);
						txData[7] = temperature+40;
						CanTX(CORE_BROADCAST_STATUS, 8);
					}
					else if (counter == 1)
					{
						int numMessages = 2;
						if (IS_BMS12I) numMessages = 3;
						for (int packet=0; packet<numMessages; packet++)
						{
							for (int n=0; n<4; n++)
							{
								txData[n*2] = voltage[packet*4+n]>>8; // Top 8 bits
								txData[n*2+1] = voltage[packet*4+n]&0xFF; // Bottom 8 bits
							}
							CanTX(BMS_BASE_ID+packet+1, 8);
						}
					}
					else if (counter == 2 && !IS_BMS12I)
					{
						for (int packet=0; packet<2; packet++)
						{
							for (int n=0; n<4; n++)
							{
								txData[n*2] = voltage[8+packet*4+n]>>8; // Top 8 bits
								txData[n*2+1] = voltage[8+packet*4+n]&0xFF; // Bottom 8 bits
							}
							CanTX(BMS_BASE_ID+10+packet+1, 8);
						}
					}
					else if (counter == 3)
					{
						if (settings[SHUNT_SIZE] != NO_SHUNT) // If it's set to "no shunt", could have CAN current sensor
						{
							long scaledValue = 8388608L + current;
							txData[0] = scaledValue>>16;
							txData[1] = scaledValue>>8;
							txData[2] = scaledValue;
							for (int n=3; n<8; n++) txData[n] = 0; // Zero for other bytes
							CanTX(CAN_CURRENT_SENSOR_ID, 8);
						}

						// TC Charger CAN stuff. Voltage stored in first byte, sent to charger in 0.1V resolution
						int targetVoltage = (short)settings[CHARGER_VOLTAGE]*10;
						// Current stored in bottom 7 bits of second byte, converted to 0.1A resolution
						int currentLimit = (short)(settings[CHARGER_CURRENT]&0b01111111)*10;

						txData[0] = targetVoltage>>8;
						txData[1] = targetVoltage%256;
						txData[2] = currentLimit>>8;
						txData[3] = currentLimit%256;
						txData[4] = !hvRelayEnabled;		
						txData[5] = txData[6] = txData[7] = 0; // Reserved, zero
						CanTX(TC_CHARGER1_RX_ID, 8); // Send control packet to TC Charger				
					}
				}

				if (SMA_INTEGRATION)
				{
					smaPacket++;
					if (smaPacket >= 8) smaPacket = 0;

					for (int n=0; n<8; n++) txData[n] = 0;

					if (smaPacket == 0) // Frame 0x351 has voltage and current limits
					{
						unsigned short maxPackVoltage = (unsigned short)settings[CHARGER_VOLTAGE]*10;
						txData[0] = maxPackVoltage&0xFF;
						txData[1] = maxPackVoltage>>8;
						unsigned short maxChargeCurrent = (unsigned short)settings[CHARGER_CURRENT]*10; // 1280 max value I think?

						// [April 2019] Function to taper current as it approaches full
						int packVoltageInTenths = (packVoltage/100L);
						
						if (packVoltageInTenths >= maxPackVoltage)
							maxChargeCurrent = 0;
						else if (packVoltageInTenths > maxPackVoltage-20) // Within 2V of max
							maxChargeCurrent = maxChargeCurrent * (maxPackVoltage-packVoltageInTenths)/20;

						unsigned short maxDischargeCurrent = 5000; // 500 amps, hardcoded, should be enough

						// Sept 2019: Goodwe inverters may not be responding to error, so limit max current if any over-voltage cells
						if (hvRelayTimer > 0) maxChargeCurrent = 0;
						if (lvRelayTimer > 0) maxDischargeCurrent = 0;
						
						txData[2] = maxChargeCurrent&0xFF;
						txData[3] = maxChargeCurrent>>8;
						txData[4] = maxDischargeCurrent&0xFF;
						txData[5] = maxDischargeCurrent>>8;
						unsigned short minPackVoltage = (unsigned short)settings[NUM_CELLS]*(150+settings[BMS_MIN_VOLTAGE])/10;
						txData[6] = minPackVoltage&0xFF;
						txData[7] = minPackVoltage>>8;
						CanTX(0x351, 8);
					}
					else if (smaPacket == 1) // Frame 0x355 has SoC etc
					{
						txData[0] = soc;
						txData[2] = 100; // SoH
						txData[4] = (soc*100)&0xFF;
						txData[5] = (soc*100)>>8;
						CanTX(0x355, 8);
					}
					else if (smaPacket == 2) // Frame 0x356 has battery volts, current and temp
					{
						short battVoltsInHundredths = packVoltage/10L;
						txData[0] = battVoltsInHundredths & 0xFF;
						txData[1] = battVoltsInHundredths>>8;

						short battAmpsInTenths = current/100L;
						txData[2] = battAmpsInTenths & 0xFF;
						txData[3] = battAmpsInTenths>>8;

						short battTempInTenths = 250; // 25degC by default
						if (temperature > -40) battTempInTenths = temperature*10; // Only if sensor present
						txData[4] = battTempInTenths&0xFF;
						txData[5] = battTempInTenths>>8;
						txData[6] = txData[7] = 0;
						CanTX(0x356, 8);
					}
					else if (smaPacket == 3) // Frame 0x35A has any errors
					{
						if (hvRelayEnabled)
							txData[0] |= 0b00001000; // Bit 3 = overvoltage clear
						else
							txData[0] |= 0b00000100; // Bit 2 = overvoltage begin
						if (lvRelayEnabled)
							txData[0] |= 0b00100000; // Bit 5 = undervoltage clear
						else
							txData[0] |= 0b00010000; // Bit 4 = undervoltage begin
						if (thermalTrip)
							txData[0] |= 0b01000000; // Bit 6 = overtemperature begin
						else
							txData[0] |= 0b10000000; // Bit 7 = overtemp clear
						if (overcurrentTrip)
							txData[1] |= 0b01000000; // Byte 1, bit 6: Overcurrent begin
						else
							txData[1] |= 0b10000000; // Byte 1, bit 7: Overcurrent clear
			
						CanTX(0x35A, 8);
					}
					else if (smaPacket == 4) // This one is for Pylontech / Luxpower, but SMAs won't mind it being there too
					{
						if (!hvRelayEnabled) txData[0] |= 0b00000010; // Bytes 0 and 1 have protection flags
						if (!lvRelayEnabled) txData[0] |= 0b00000100;
						if (thermalTrip)      txData[0] |= 0b00001000;
						if (tooColdToCharge)  txData[0] |= 0b00010000;
						if (overcurrentTrip) {txData[0] |= 0b10000000;
											  txData[1] |= 0b00000001; } // Signal both discharge and charge overcurrent at once
						// Bytes 2 and 3 have warnings instead of errors - can hopefully ignore
						txData[4] = 1; // Pack number
						txData[5] = 'P'; txData[6] = 'N'; // From Pylontech datasheet, might be manufacturer code?
						CanTX(0x359, 8);
					}
					else if (smaPacket == 5) // Also Pylontech / Luxpower
					{
						if (hvRelayEnabled) txData[0] |= 0b10000000; // Charge enable bit
						if (lvRelayEnabled) txData[0] |= 0b01000000; // Discharge enable bit
						CanTX(0x35C, 2);
					}
					else if (smaPacket == 6) // Frame 0x35E has manufacturer description, 8 ASCII characters
					{
						txData[0] = 'Z'; txData[1] = 'E'; txData[2] = 'V'; txData[3] = 'A';
						txData[4] = ' '; txData[5] = 'B'; txData[6] = 'M'; txData[7] = 'S';
						CanTX(0x35E, 8);
					}
				}

				if (SOLAX_INTEGRATION)
				{
					solaxPacket++;
					if (solaxPacket >= 10) solaxPacket = 0;

					// Messages to Solax need to be sent at least 100ms apart (so, every 2 counts), and all within 1sec or faster
					if (solaxPacket == 1)
					{
						unsigned short maxPackVoltage = (unsigned short)settings[CHARGER_VOLTAGE]*10;
						txData[0] = maxPackVoltage&0xFF;
						txData[1] = maxPackVoltage>>8;
						unsigned short minPackVoltage = (unsigned short)settings[NUM_CELLS]*(160+settings[BMS_MIN_VOLTAGE])/10;
						txData[2] = minPackVoltage&0xFF;
						txData[3] = minPackVoltage>>8;
						unsigned short maxChargeCurrent = (unsigned short)settings[CHARGER_CURRENT]*10;
						txData[4] = maxChargeCurrent&0xFF;
						txData[5] = maxChargeCurrent>>8;
						unsigned short maxDischargeCurrent = 5000; // 500 amps, hardcoded, should be enough
						txData[6] = maxDischargeCurrent&0xFF;
						txData[7] = maxDischargeCurrent>>8;
						CanTX(0x1872, 8);
					}
					else if (solaxPacket == 3) // 2 counts at 16hz means 125ms since last msg
					{
						unsigned short maxTemp = Cap((short)(settings[BMS_MAX_TEMP]-40)*10, 0, 100);
						txData[0] = maxTemp&0xFF;
						txData[1] = maxTemp>>8;
						unsigned short minTemp = Cap((short)(settings[BMS_MIN_TEMP]-40)*10, 0, 100);
						txData[2] = minTemp&0xFF;
						txData[3] = minTemp>>8;
						txData[4] = soc;
						txData[5] = 0;
						long tenthsOfAmpHours = charge / 3600L;
						long tenthsOfVolts = packVoltage / 100L;
						unsigned short tensOfWattHours = tenthsOfAmpHours * tenthsOfVolts / 1000L;
						txData[6] = tensOfWattHours&0xFF;
						txData[7] = tensOfWattHours>>8;
						CanTX(0x1873, 8);
					}
					else if (solaxPacket == 5)
					{
						for (int n=0; n<8; n++) txData[n] = 0;
						if (!hvRelayEnabled) txData[2] = 1;
						if (!lvRelayEnabled) txData[6] = 1; // Disable bits
						CanTX(0x1874, 8);
					}
					else if (solaxPacket == 7)
					{
						for (int n=0; n<8; n++) txData[n] = 0;
						unsigned char flagBits;
						if (!hvRelayEnabled) flagBits |= 0b00000100;
						if (!lvRelayEnabled) flagBits |= 0b00001000;
						if (overcurrentTrip) flagBits |= 0b00100000;
						if (error == OVERTEMP_ERROR) flagBits |= 0b01000000;
						if (error == BMS_UNDERTEMP) flagBits |= 0b10000000;
						CanTX(0x1875, 8);
					}
					else if (solaxPacket == 9)
					{
						short battVoltsInTenths = packVoltage/100L;
						txData[0] = battVoltsInTenths & 0xFF;
						txData[1] = battVoltsInTenths>>8;
						short battAmpsInTenths = current/100L;
						txData[2] = battAmpsInTenths & 0xFF;
						txData[3] = battAmpsInTenths>>8;
						txData[4] = 1; // Subsystem number - always 1 will do?
						txData[5] = txData[6] = txData[7] = 0;
						CanTX(0x1876, 8);
					}
				}
			}

			counter++;
		} // End fast loop		
				
		if (counter >= 4) // Slow loop, about 4Hz
		{
			counter = 0;
			slowCounter++;
			if (slowCounter >= 8) slowCounter = 0;

			char notAllZeroVolts = false;
			char hvTrip = false;
			char hvReset = true;
			char lvTrip = false;
			char lvReset = true;
			char criticalVoltage = false;
			packVoltage = 0;
			int minCellVoltage = 5000;
			int maxCellVoltage = 0;
			for (int n=0; n<settings[NUM_CELLS]; n++) // Calculate average voltage over last 4 samples, and update shunts if required
			{
				int average = 0;
				for (int c=0; c<4; c++) average += voltages[n][c];
				
				int correction = LOWER_LTC_CORRECTION;
				if (n>=8 && !IS_BMS12I) correction = UPPER_LTC_CORRECTION; // Separate calibration for 2nd LTC in BMS16
				voltage[n] = average/4 + correction; // Added due to voltage drop across 3.3K input resistors.
				if (n==0 || (n==8 && !IS_BMS12I)) voltage[n] -= correction/2; // First and 9th cells only have single 3.3Kohm resistor drop

				if (voltage[n] < minCellVoltage) minCellVoltage = voltage[n];
				if (voltage[n] > maxCellVoltage) maxCellVoltage = voltage[n];

				packVoltage += voltage[n];

				if (voltage[n] > 0) notAllZeroVolts = true;

				int hysteresis = 0;
				if (settings[STATIONARY_VERSION]) hysteresis = settings[BMS_HYSTERESIS]; // Only apply hysteresis in stationary mode

				// [April 2019] Changed hysteresis to only affect reset threshold, not trip (safer this way)
				if (voltage[n] < (150+settings[BMS_MIN_VOLTAGE])*10) lvTrip = true;
				if (voltage[n] < (150+settings[BMS_MIN_VOLTAGE]+hysteresis)*10) lvReset = false;

				if (voltage[n] > (200+settings[BMS_MAX_VOLTAGE])*10) hvTrip = true;
				if (voltage[n] > (200+settings[BMS_MAX_VOLTAGE]-hysteresis)*10) hvReset = false;

				if (voltage[n] < POWER_OFF_MILLIVOLTS && voltage[n] > 100) criticalVoltage = true;
					// <100 might indicate zero reading, could be incorrect programming, need to be able to change without shutting down
			}
			midpointCellVoltage = (minCellVoltage + maxCellVoltage)/2;
			
			if (canCurrentSensorTimeout > 0)
				canCurrentSensorTimeout--; // We have recently received current data over CAN; decrement timeout
			else
			{	// No recent CAN current data, use internal current sensor if available
				current = 0;
				for (int n=0; n<4; n++)
				{
					current += currentBuffer[n];
					currentBuffer[n] = 0;	
				}
				current /= 32L;
			}

			charge -= current/40L;
			if (charge < 0) charge = 0;
			if (charge > maxCharge) charge = maxCharge;
			if (current < 200 && current > -200) current = 0; // Round off if less than 0.2A, for tidier display
			
			// Reset/resynchronise SoC to 100% if pack voltage is full and low current flowing
			if (packVoltage >= (unsigned int)settings[FULL_VOLTAGE]*1000L && (current < 2000 && current > -2000))
				charge = maxCharge;

			short oldsoc = soc;
			soc = (201L*(charge>>8)/(maxCharge>>8))>>1; // Effectively *100.5 so we get rounding instead of truncating
			if (oldsoc > settings[SOC_WARNING] && soc <= settings[SOC_WARNING]) error = LOW_SOC_ERROR;

			if (lvReset)
				lvRelayTimer = 0;
			else if ((lvTrip || lvRelayTimer > 0) && lvRelayTimer < 100) // lvTrip can kick off timer, always counts until lvReset condition
				lvRelayTimer++;
			
			if (hvReset)
				hvRelayTimer = 0;
			else if ((hvTrip || hvRelayTimer > 0) && hvRelayTimer < 100)
				hvRelayTimer++;

			if (criticalVoltage)
				criticalVoltageTimer++;
			else
				criticalVoltageTimer = 0;

			if (criticalVoltageTimer > 4 && criticalVoltageTimer < 40) error = BMS_LOW_WARNING; // Warn of critical voltage
				// Need to stop triggering warning after 10s so people can hold power button and enter setup to fix cell num
			if (criticalVoltageTimer > 40)
			{
				PutLTCsToSleep();
				PORTD &= ~PWR_HOLD; // Then shut down after 10 seconds
			}
			if (settings[CURRENT_WARNING] < 121 && Abs(current/1000L) > settings[CURRENT_WARNING]*10
				&& !(errorAckFlags&(1L<<OVERCURRENT_WARNING)))
				error = OVERCURRENT_WARNING;

			if (settings[CURRENT_TRIP] < 121 && Abs(current/1000L) > settings[CURRENT_TRIP]*10)
			{
				if (overcurrentTimer < 100) overcurrentTimer++;
			}
			else
				overcurrentTimer = 0;

			if (overcurrentTimer > 4)
			{
				if (!overcurrentTrip) error = OVERCURRENT_SHUTDOWN; // Only raise error when condition first arrives so it can be reset
				overcurrentTrip = true;
			}

			if (settings[BMS_MAX_TEMP] < 141 && temperature+40 > settings[BMS_MAX_TEMP])
			{
				if (!thermalTrip) error = OVERTEMP_ERROR; // Only raise error when condition first arrives, so it can be reset
				thermalTrip = true;
			}
			else if (settings[BMS_MAX_TEMP] == 141 || temperature+40 < settings[BMS_MAX_TEMP]-10) // 10 degrees hysteresis
				thermalTrip = false;

			if (temperature > -40 && temperature+40 < settings[BMS_MIN_TEMP] && settings[BMS_MIN_TEMP] > 0)
				tooColdToCharge = true;
			else if (temperature == -40 || temperature+40 > settings[BMS_MIN_TEMP] || settings[BMS_MIN_TEMP] == 0)
				tooColdToCharge = false;

			if (overcurrentTrip || thermalTrip || sleeping)
			{
				if (INVERT_SSRS)
					RELAY_PORT |= LV_RELAY + HV_RELAY;
				else
					RELAY_PORT &= ~(LV_RELAY + HV_RELAY);
				lvRelayEnabled = hvRelayEnabled = false;
			}
			else
			{
				if (lvRelayTimer > 4) // About 1 second
				{
					if (INVERT_SSRS)
						RELAY_PORT |= LV_RELAY;
					else
						RELAY_PORT &= ~LV_RELAY;
					lvRelayEnabled = false;
					if (error == NO_ERROR && !(errorAckFlags&(1L<<BMS_LOW_WARNING))) error = BMS_LOW_WARNING;
				}
				else if (settings[STATIONARY_VERSION] == true || justWokeUp)
				{
					if (INVERT_SSRS)
						RELAY_PORT &= ~LV_RELAY;
					else
						RELAY_PORT |= LV_RELAY;
					lvRelayEnabled = true;
					if (error == BMS_LOW_WARNING) error = NO_ERROR;
				}

				if (hvRelayTimer > 4 || tooColdToCharge) // About 1 second
				{
					if (INVERT_SSRS)
						RELAY_PORT |= HV_RELAY;
					else
						RELAY_PORT &= ~HV_RELAY;
					hvRelayEnabled = false;
					if (error == NO_ERROR)
					{
						if (tooColdToCharge && !(errorAckFlags&(1L<<BMS_UNDERTEMP)))
							error = BMS_UNDERTEMP;
						else if (!(errorAckFlags&(1L<<BMS_HIGH_WARNING)))
							error = BMS_HIGH_WARNING;
					}
				}
				else if (settings[STATIONARY_VERSION] == true || justWokeUp)
				{
					if (INVERT_SSRS)
						RELAY_PORT &= ~HV_RELAY;
					else
						RELAY_PORT |= HV_RELAY;
					hvRelayEnabled = true;
					if (error == BMS_HIGH_WARNING) error = NO_ERROR;
				}
			}

			justWokeUp = false;

			if (sleeping) // Blink green
			{
				LED_PORT &= ~(RED+GREEN);
				if (slowCounter == 0) LED_PORT |= GREEN;
			}
			else if (thermalTrip)
			{
				LED_PORT &= ~(RED+GREEN);
				if (slowCounter & 0x02)
					LED_PORT |= RED;
				else
					LED_PORT |= GREEN;
			}
			else if (lvRelayEnabled && hvRelayEnabled) // All happy = green LED
			{
				LED_PORT &= ~RED;
				LED_PORT |= GREEN;
			}
			else
			{
				if (!lvRelayEnabled) // Undervoltage = Blink red
				{
					LED_PORT &= ~(GREEN+RED);
					if (slowCounter & 0x02)
						LED_PORT |= RED;
				}
				else // Overvoltage = Solid red
				{
					LED_PORT |= RED;
					LED_PORT &= ~GREEN;
				}
			}
		}
	}
}

void CanTX(long packetID, char bytes)
{
	CANPAGE = 0x00; // Select MOB0 for transmission
	while (CANEN2 & (1<<ENMOB0)); // Wait for MOB0 to be free
	CANSTMOB = 0x00;

	if (USE_29BIT_IDS) // CAN 2.0b is 29-bit IDs, CANIDT4 has bits 0-4 in top 5 bits, CANID3 has 5-12
	{
		CANIDT1 = packetID>>21;
		CANIDT2 = packetID>>13;
		CANIDT3 = packetID>>5;
		CANIDT4 = (packetID & 0b00011111)<<3;
	}
	else // CAN 2.0a is 11-bit IDs, IDT1 has top 8 bits, IDT2 has bottom three bits BUT at top of byte!
	{
		CANIDT1 = (packetID>>3); // Packet ID
		CANIDT2 = (packetID & 0x07)<<5;
		CANIDT3 = 0x00;
		CANIDT4 = 0x00;
	}

	for (int8_t i=0; i<bytes; i++) CANMSG = txData[i];

	CANCDMOB = (1<<CONMOB0) | (bytes<<DLC0) | ((1<<IDE)*USE_29BIT_IDS); // Enable transmission, 8-bit data

	char timeout = 0;
	while (!(CANSTMOB & (1<<TXOK)) && !(CANSTMOB & (1<<AERR))) // Wait for transmission to finish (via setting of TXOK flag)
	{
		_delay_us(100);
		if (timeout++ > 100) break;
	}

	CANCDMOB = 0x00; // Disable transmission
	CANSTMOB = 0x00; // Clear TXOK flag
}

void SetupPorts()
{
	DDRB = 0b01111111; // PB0 = CAN enable, PB1 = SCKI2, PB2 = CSBI2, PB3 = CSBI, PB4 = SDI, PB5 = SCKI, PB6 = 12V enable
	DDRC = 0b00000000; // PC6 = SDO
	DDRD = 0b01111101; // PD0 = Pwr Hold, PD2 = Red, PD3 = Green, PD4 = SDI2, PD5 = LVout, PD6 = HVout, PD7 = SDO2

	PORTC |= SDO; // Pull up, for PCB version 2.1 that doesn't have the lower ADUM1401

	PORTC |= (1<<PC0); // Pull-up on sleep pin	

	// Timer 0 used for system clock
	TCCR0B = (1<<CS00); // Clock on, 1:1 prescaler
	TIMSK0 |= (1<<TOIE0);

	// ADC stuff
	ADMUX = ADC_VREF_TYPE;
	ADCSRA = 0b10000110; // ADEN plus 110 prescaler = /64 for 125khz ADC clock
	ADCSRB |= (1<<AREFEN);
	AMP1CSR |= (1<<AMP1EN) + (1<<AMP1G1); // 20x gain on differential ADC
	
	// External interrupt for sleep pin
	EICRA = 0b01000000; // 01 for INT3 means interrupt on any logical change
	EIMSK = 0b00001000; // Enables INT3 only

	// DAC stuff
	DACON = 0b00000011; // DAC enable, and D2A output pin enable
	DAC = 511;

	// CAN init stuff
	CANGCON = (1<<SWRES); // Software reset
	CANTCON = 0; // CAN timing prescaler set to 0
	if (CAN_BAUD == 1000)
		CANBT1 = 0x00;
	else if (CAN_BAUD == 500)
		CANBT1 = 0x02;
	else if (CAN_BAUD == 250)
		CANBT1 = 0x06;
	else
		CANBT1 = 0x0E;
	CANBT2 = 0x04;
	CANBT3 = 0x13; // These set baud rate (see ATmega16M1 manual, page 203)
	if (CAN_BAUD == 1000) CANBT3 = 0x12;

	for (int8_t mob=0; mob<6; mob++)
	{
		CANPAGE = (mob<<4); // Select MOB 0-5
		CANCDMOB = 0x00; // Disable MOB
		CANSTMOB = 0x00; // Clear MOB status register
	}

	CANPAGE = (1<<MOBNB0); // Select MOB1
	CANIE2 = (1<<IEMOB1); // Enable interrupts on MOB1 for reception and transmission
	CANGIE = (1<<ENIT) | (1<<ENRX); // Enable interrupts on receive and transmit
	CANIDM1 = CANIDM2 = CANIDM3 = CANIDM4 = 0x00; // CAN ID mask, zero will let all IDs pass
	CANCDMOB = (1<<CONMOB1) + (8<<DLC0) + ((1<<IDE)*USE_29BIT_IDS); // Enable reception, 8-bit data length
	CANGCON |= (1<<ENASTB); // Enable mode. CAN channel enters enable mode after 11 recessive bits have been read
}

void WriteSPIByte(unsigned char byte)
{
	unsigned short mask = 0b10000000; // MSB first
	for (int n=0; n<8; n++) // 8 bits = 1 byte
	{
		// Prepare pin
		if ((byte & mask) > 0) // Bit is on
			SDI_PORT |= SDI;
		else
			SDI_PORT &= ~SDI;

		_delay_us(2);
		SCKI_PORT |= SCKI; // Clock goes high = register SDI state
		_delay_us(2);
		SCKI_PORT &= ~SCKI; // Clock goes low
		mask = mask>>1;
	}
}

unsigned char ReadSPIByte()
{
	unsigned char byte = 0;
	unsigned char mask = 0b10000000;
	for (int n=0; n<8; n++)
	{
		_delay_us(2);
		SCKI_PORT |= SCKI; // Clock goes high = register SDI state
		_delay_us(2);
		if (SDO_PORT & SDO) byte += mask;
		SCKI_PORT &= ~SCKI; // Clock goes low
		mask = mask>>1;
	}
	return byte;
}

void WriteSPIByte2(unsigned char byte)
{
	unsigned short mask = 0b10000000; // MSB first
	for (int n=0; n<8; n++) // 8 bits = 1 byte
	{
		// Prepare pin
		if ((byte & mask) > 0) // Bit is on
			SDI2_PORT |= SDI2;
		else
			SDI2_PORT &= ~SDI2;

		_delay_us(2);
		SCKI2_PORT |= SCKI2; // Clock goes high = register SDI state
		_delay_us(2);
		SCKI2_PORT &= ~SCKI2; // Clock goes low
		mask = mask>>1;
	}
}

unsigned char ReadSPIByte2()
{
	unsigned char byte = 0;
	unsigned char mask = 0b10000000;
	for (int n=0; n<8; n++)
	{
		_delay_us(2);
		SCKI2_PORT |= SCKI2; // Clock goes high = register SDI state
		_delay_us(2);
		if (SDO2_PORT & SDO2) byte += mask;
		SCKI2_PORT &= ~SCKI2; // Clock goes low
		mask = mask>>1;
	}
	return byte;
}

char LoadSettingsFromEEPROM()
{
	unsigned char tempSettings[NUM_SETTINGS];
	
	if (eeprom_read_dword((uint32_t*)(EEPROM_OFFSET + NUM_SETTINGS + 6)) != 0xC0FFEE)
		return EEPROM_BLANK;
	
	unsigned short checksum = 0;
	for (int i=0; i<NUM_SETTINGS; i++)
	{
		tempSettings[i] = eeprom_read_byte((uint8_t*)(EEPROM_OFFSET + i));
		checksum += tempSettings[i];
	}

	shuntCalibrationCoarse = eeprom_read_word((uint16_t*)(EEPROM_OFFSET+NUM_SETTINGS));
	shuntCalibrationFine = eeprom_read_word((uint16_t*)(EEPROM_OFFSET+NUM_SETTINGS+2));
	checksum += shuntCalibrationCoarse + shuntCalibrationFine;

	if (checksum != eeprom_read_word((uint16_t*)(EEPROM_OFFSET + NUM_SETTINGS + 4))) return EEPROM_CORRUPT;

	// Otherwise, settings are correct and we can copy to the real thing
	for (int i=0; i<NUM_SETTINGS; i++) settings[i] = tempSettings[i];
	
    eeprom_write_byte(0, 0); // EEAR = 0; // Park EEPROM pointer at sacrificial location 0

	return EEPROM_CORRECT;
}

void SaveSettingsToEEPROM()
{
	wdt_disable();	// Disable watchdog
	cli();			// Disable interrupts

	unsigned short checksum = 0;
	for (int i=0; i<NUM_SETTINGS; i++)
	{
		eeprom_write_byte((uint8_t*)(EEPROM_OFFSET + i), settings[i]);
		checksum += settings[i];
	}

	checksum += shuntCalibrationCoarse + shuntCalibrationFine;

	eeprom_write_word((uint16_t*)(EEPROM_OFFSET + NUM_SETTINGS), shuntCalibrationCoarse);
	eeprom_write_word((uint16_t*)(EEPROM_OFFSET + NUM_SETTINGS + 2), shuntCalibrationFine);
	eeprom_write_word((uint16_t*)(EEPROM_OFFSET + NUM_SETTINGS + 4), checksum); // Add checksum
	eeprom_write_dword((uint32_t*)(EEPROM_OFFSET + NUM_SETTINGS + 6), 0xC0FFEE); // Add watermark

	sei();
	wdt_enable(WDTO_250MS);

	EEAR = 0;
}

long GetCurrent()
{
	if (settings[SHUNT_SIZE] == NO_SHUNT) return 0;
	
	long sample = (ReadADC(POS) - ReadADC(NEG))*10L;
	sample -= shuntCalibrationCoarse;

	if (sample < 400 && sample > -400) // If sample is small, use amplified version
	{
		sample = ReadADC(15);
		if (sample > 511) sample -= 1024; // Make it signed
		sample -= shuntCalibrationFine;
	}

	// Default shunt is 75mV per 100A
	// AD8218 has 20x gain, so 15mV per 1A
	// ADCs are 4.88mV per ADC, 3.07 ADCs per amp
	// Amplified one is 61.5 ADCs per amp, or 0.0615 ADCs per milliamp
	// Can multiply by 16.25 to get quite accurate milliamps

	int multiplier = 1L;
	if (settings[SHUNT_SIZE] == SHUNT_200A) multiplier = 2L;
	else if (settings[SHUNT_SIZE] == SHUNT_500A)
	{
		if (USE_600A_SHUNT) multiplier = 6L;
		else multiplier = 5L;
	}

	long current = (sample*65L/2L)*SHUNT_CALIBRATION*multiplier / 100L;
	
	// V3 uses AMC1200B shunt amplifier chip, which has 8x gain instead of 20x for AD8218
	if (BMS_VERSION >= 3) current = current*5L/2L;

	current += CURRENT_OFFSET*multiplier;

	// current = current*3L/2L; // Uncomment this to custom scale for 50mV shunts

	if (current < 200L*multiplier && current > -200L*multiplier) current = 0; // Deadband noise around zero point

	return current;
}

int GetTemperature()
{
	// ADC vals for temps:        -40, -20    0   20   40   60   80 100 120 140 160 dec C
	static short tempData[11] = { 990, 950, 783, 568, 356, 204, 114, 65, 38, 23, 15 };

	int tempADC = ReadADC(TEMP);
	tempBuffer = 0;
	temperature = -40; // i.e "not available" if the algorithm below doesn't find it (-20 to 160degC)
	for (int n=0; n<10; n++)
	{
		if (tempADC <= tempData[n] && tempADC > tempData[n+1]) // we're between samples  
		{
			temperature = -40 + n*20 + 20*(tempADC - tempData[n])/(tempData[n+1] - tempData[n]);
			break;
		}
	}

	// if (temperature < -10) temperature = -40; // Occasionally needed if temp input has low resistance for some reason

	return temperature;
}
